@@ -67,8 +67,8 @@ static void * load_library(const char * rel_path, const char * src_dir, int err)
 
 #if defined(_OS_WINDOWS_)
 #define PATH_EXISTS() win_file_exists(wpath)
-    wchar_t wpath[2*JL_PATH_MAX + 1] = {0};
-    if (!utf8_to_wchar(path, wpath, 2*JL_PATH_MAX)) {
+    wchar_t *wpath = utf8_to_wchar(path);
+    if (!wpath) {
         jl_loader_print_stderr3("ERROR: Unable to convert path ", path, " to wide string!\n");
         exit(1);
     }
@@ -77,9 +77,21 @@ static void * load_library(const char * rel_path, const char * src_dir, int err)
 #define PATH_EXISTS() !access(path, F_OK)
     handle = dlopen(path, RTLD_NOW | (err ? RTLD_GLOBAL : RTLD_LOCAL));
 #endif
-    if (handle == NULL) {
-        if (!err && !PATH_EXISTS())
+    if (handle != NULL) {
+#if defined(_OS_WINDOWS_)
+        free(wpath);
+#endif
+    }
+    else {
+        if (!err && !PATH_EXISTS()) {
+#if defined(_OS_WINDOWS_)
+            free(wpath);
+#endif
             return NULL;
+        }
+#if defined(_OS_WINDOWS_)
+        free(wpath);
+#endif
 #undef PATH_EXISTS
         jl_loader_print_stderr3("ERROR: Unable to load dependent library ", path, "\n");
 #if defined(_OS_WINDOWS_)
@@ -91,9 +103,9 @@ static void * load_library(const char * rel_path, const char * src_dir, int err)
                        NULL, GetLastError(),
                        MAKELANGID(LANG_ENGLISH, SUBLANG_ENGLISH_US),
                        (LPWSTR)&wmsg, 0, NULL);
-        char err[256] = {0};
-        wchar_to_utf8(wmsg, err, 255);
-        jl_loader_print_stderr3("Message:", err, "\n");
+        char *errmsg = wchar_to_utf8(wmsg);
+        jl_loader_print_stderr3("Message:", errmsg, "\n");
+        free(errmsg);
 #else
         char *dlerr = dlerror();
         if (dlerr != NULL) {
@@ -114,20 +126,20 @@ static void * lookup_symbol(const void * lib_handle, const char * symbol_name) {
 }
 
 // Find the location of libjulia.
-char lib_dir[JL_PATH_MAX];
+char *lib_dir = NULL;
 JL_DLLEXPORT const char * jl_get_libdir()
 {
     // Reuse the path if this is not the first call.
-    if (lib_dir[0] != 0) {
+    if (lib_dir) {
         return lib_dir;
     }
 #if defined(_OS_WINDOWS_)
     // On Windows, we use GetModuleFileNameW
-    wchar_t libjulia_path[JL_PATH_MAX];
+    wchar_t *libjulia_path = utf8_to_wchar(LIBJULIA_NAME);
     HMODULE libjulia = NULL;
 
     // Get a handle to libjulia.
-    if (!utf8_to_wchar(LIBJULIA_NAME, libjulia_path, JL_PATH_MAX)) {
+    if (!libjulia_path) {
         jl_loader_print_stderr3("ERROR: Unable to convert path ", LIBJULIA_NAME, " to wide string!\n");
         exit(1);
     }
@@ -136,14 +148,18 @@ JL_DLLEXPORT const char * jl_get_libdir()
         jl_loader_print_stderr3("ERROR: Unable to load ", LIBJULIA_NAME, "!\n");
         exit(1);
     }
-    if (!GetModuleFileNameW(libjulia, libjulia_path, JL_PATH_MAX)) {
+    free(libjulia_path);
+    libjulia_path = (wchar_t*)malloc(32768 * sizeof(wchar_t)); // max long path length
+    if (!GetModuleFileNameW(libjulia, libjulia_path, 32768)) {
         jl_loader_print_stderr("ERROR: GetModuleFileName() failed\n");
         exit(1);
     }
-    if (!wchar_to_utf8(libjulia_path, lib_dir, JL_PATH_MAX)) {
+    lib_dir = wchar_to_utf8(libjulia_path);
+    if (!lib_dir) {
         jl_loader_print_stderr("ERROR: Unable to convert julia path to UTF-8\n");
         exit(1);
     }
+    free(libjulia_path);
 #else
     // On all other platforms, use dladdr()
     Dl_info info;
@@ -155,7 +171,7 @@ JL_DLLEXPORT const char * jl_get_libdir()
         }
         exit(1);
     }
-    strcpy(lib_dir, info.dli_fname);
+    lib_dir = strdup(info.dli_fname);
 #endif
     // Finally, convert to dirname
     const char * new_dir = dirname(lib_dir);
@@ -211,13 +227,13 @@ static void read_wrapper(int fd, char **ret, size_t *ret_len)
     size_t have_read = 0;
     while (1) {
         ssize_t n = read(fd, buf + have_read, len - have_read);
-        have_read += n;
         if (n == 0) break;
         if (n == -1 && errno != EINTR) {
             perror("(julia) libstdcxxprobe read");
             exit(1);
         }
         if (n == -1 && errno == EINTR) continue;
+        have_read += n;
         if (have_read == len) {
             buf = (char *)realloc(buf, 1 + (len *= 2));
             if (!buf) {
@@ -339,6 +355,17 @@ static char *libstdcxxprobe(void)
 void *libjulia_internal = NULL;
 void *libjulia_codegen = NULL;
 __attribute__((constructor)) void jl_load_libjulia_internal(void) {
+#if defined(_OS_LINUX_)
+    // Julia uses `sigwait()` to handle signals, and all threads are required
+    // to mask the corresponding handlers so that the signals can be waited on.
+    // Here, we setup that masking early, so that it is inherited by any threads
+    // spawned (e.g. by constructors) when loading deps of libjulia-internal.
+
+    sigset_t all_signals, prev_mask;
+    sigfillset(&all_signals);
+    pthread_sigmask(SIG_BLOCK, &all_signals, &prev_mask);
+#endif
+
     // Only initialize this once
     if (libjulia_internal != NULL) {
         return;
@@ -500,13 +527,25 @@ __attribute__((constructor)) void jl_load_libjulia_internal(void) {
     }
     void *fptr = lookup_symbol(RTLD_DEFAULT, "jl_get_pgcstack_static");
     void *(*key)(void) = lookup_symbol(RTLD_DEFAULT, "jl_pgcstack_addr_static");
-    if (fptr != NULL && key != NULL)
-        jl_pgcstack_setkey(fptr, key);
+    _Atomic(char) *semaphore = lookup_symbol(RTLD_DEFAULT, "jl_pgcstack_static_semaphore");
+    if (fptr != NULL && key != NULL && semaphore != NULL) {
+        char already_used = 0;
+        atomic_compare_exchange_strong(semaphore, &already_used, 1);
+        if (already_used == 0) // RMW succeeded - we have exclusive access
+            jl_pgcstack_setkey(fptr, key);
+    }
 #endif
 
     // jl_options must be initialized very early, in case an embedder sets some
     // values there before calling jl_init
     ((void (*)(void))jl_init_options_addr)();
+
+#if defined(_OS_LINUX_)
+    // Restore the original signal mask. `jl_init()` will later setup blocking
+    // for the specific set of signals we `sigwait()` on, and any threads spawned
+    // during loading above will still retain their inherited signal mask.
+    pthread_sigmask(SIG_SETMASK, &prev_mask, NULL);
+#endif
 }
 
 // Load libjulia and run the REPL with the given arguments (in UTF-8 format)
