@@ -93,7 +93,7 @@ function printvalue(f::MbyFunc, io::IO, value::TOMLValue)
     value isa Dates.Time     ? Base.print(io, Dates.format(value, Dates.dateformat"HH:MM:SS.sss")) :
     value isa Dates.Date     ? Base.print(io, Dates.format(value, Dates.dateformat"YYYY-mm-dd")) :
     value isa Bool           ? Base.print(io, value ? "true" : "false") :
-    value isa Integer        ? Base.print(io, Int64(value)) :  # TOML specifies 64-bit signed long range for integer
+    value isa Integer        ? print_integer(io, value) :  # Julia's own printing should be compatible with TOML on integers
     value isa AbstractFloat  ? Base.print(io, isnan(value) ? "nan" :
                                               isinf(value) ? string(value > 0 ? "+" : "-", "inf") :
                                               Float64(value)) :  # TOML specifies IEEE 754 binary64 for float
@@ -104,6 +104,14 @@ function printvalue(f::MbyFunc, io::IO, value::TOMLValue)
     error("internal error in TOML printing, unhandled value")
 end
 
+function print_integer(io::IO, value::Integer)
+    value isa Signed && return Base.show(io, value)
+    # unsigned integers are printed as hex
+    n = 2 * ndigits(value, base=256)
+    Base.print(io, "0x", string(value, base=16, pad=n))
+    return
+end
+
 function print_inline_table(f::MbyFunc, io::IO, value::AbstractDict)
     Base.print(io, "{")
     for (i, (k,v)) in enumerate(value)
